/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.hadoop.record.compiler;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

/**
 *
 * @author Milind Bhandarkar
 */
public class JRecord extends JCompType {

    private String mFQName;
    private String mName;
    private String mModule;
    private ArrayList mFields;
    
    /**
     * Creates a new instance of JRecord
     */
    public JRecord(String name, ArrayList flist) {
        super(name.replaceAll("\\.","::"), name, "Record", name);
        mFQName = name;
        int idx = name.lastIndexOf('.');
        mName = name.substring(idx+1);
        mModule = name.substring(0, idx);
        mFields = flist;
    }
    
    public String getName() {
        return mName;
    }
    
    public String getJavaFQName() {
        return mFQName;
    }
    
    public String getCppFQName() {
        return mFQName.replaceAll("\\.", "::");
    }
    
    public String getJavaPackage() {
        return mModule;
    }
    
    public String getCppNameSpace() {
        return mModule.replaceAll("\\.", "::");
    }
    
    public ArrayList getFields() {
        return mFields;
    }
    
    public String getSignature() {
        StringBuffer sb = new StringBuffer();
        sb.append("L").append(mName).append("(");
        for (Iterator i = mFields.iterator(); i.hasNext();) {
            String s = ((JField) i.next()).getSignature();
            sb.append(s);
        }
        sb.append(")");
        return sb.toString();
    }
    
    public String genCppDecl(String fname) {
        return "  "+mName+" "+fname+";\n";
    }
    
    public String genJavaReadMethod(String fname, String tag) {
        return genJavaReadWrapper(fname, tag, false);
    }
    
    public String genJavaReadWrapper(String fname, String tag, boolean decl) {
        StringBuffer ret = new StringBuffer("");
        if (decl) {
            ret.append("    "+getJavaFQName()+" "+fname+";\n");
        }
        ret.append("    "+fname+"= new "+getJavaFQName()+"();\n");
        ret.append("    a_.readRecord("+fname+",\""+tag+"\");\n");
        return ret.toString();
    }
    
    public String genJavaWriteWrapper(String fname, String tag) {
        return "    a_.writeRecord("+fname+",\""+tag+"\");\n";
    }
    
    public String genJavaSlurpBytes(String b, String s, String l) {
      StringBuffer sb = new StringBuffer();
      sb.append("        {\n");
      sb.append("           int r = "+getJavaFQName()+
          ".Comparator.slurpRaw("+b+","+s+","+l+");\n");
      sb.append("           "+s+"+=r; "+l+"-=r;\n");
      sb.append("        }\n");
      return sb.toString();
    }
    
    public String genJavaCompareBytes() {
      StringBuffer sb = new StringBuffer();
      sb.append("        {\n");
      sb.append("           int r1 = "+getJavaFQName()+
          ".Comparator.compareRaw(b1,s1,l1,b2,s2,l2);\n");
      sb.append("           if (r1 <= 0) { return r1; }\n");
      sb.append("           s1+=r1; s2+=r1; l1-=r1; l2-=r1;\n");
      sb.append("        }\n");
      return sb.toString();
    }
    
    public void genCppCode(FileWriter hh, FileWriter cc)
        throws IOException {
        String[] ns = getCppNameSpace().split("::");
        for (int i = 0; i < ns.length; i++) {
            hh.write("namespace "+ns[i]+" {\n");
        }
        
        hh.write("class "+getName()+" : public ::hadoop::Record {\n");
        hh.write("private:\n");
        
        for (Iterator i = mFields.iterator(); i.hasNext();) {
            JField jf = (JField) i.next();
            hh.write(jf.genCppDecl());
        }
        hh.write("public:\n");
        hh.write("  virtual void serialize(::hadoop::OArchive& a_, const char* tag) const;\n");
        hh.write("  virtual void deserialize(::hadoop::IArchive& a_, const char* tag);\n");
        hh.write("  virtual const ::std::string& type() const;\n");
        hh.write("  virtual const ::std::string& signature() const;\n");
        hh.write("  virtual bool operator<(const "+getName()+"& peer_) const;\n");
        hh.write("  virtual bool operator==(const "+getName()+"& peer_) const;\n");
        hh.write("  virtual ~"+getName()+"() {};\n");
        int fIdx = 0;
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            hh.write(jf.genCppGetSet(fIdx));
        }
        hh.write("}; // end record "+getName()+"\n");
        for (int i=ns.length-1; i>=0; i--) {
            hh.write("} // end namespace "+ns[i]+"\n");
        }
        cc.write("void "+getCppFQName()+"::serialize(::hadoop::OArchive& a_, const char* tag) const {\n");
        cc.write("  a_.startRecord(*this,tag);\n");
        fIdx = 0;
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            String name = jf.getName();
            if (jf.getType() instanceof JBuffer) {
                cc.write("  a_.serialize("+name+","+name+".length(),\""+jf.getTag()+"\");\n");
            } else {
                cc.write("  a_.serialize("+name+",\""+jf.getTag()+"\");\n");
            }
        }
        cc.write("  a_.endRecord(*this,tag);\n");
        cc.write("  return;\n");
        cc.write("}\n");
        
        cc.write("void "+getCppFQName()+"::deserialize(::hadoop::IArchive& a_, const char* tag) {\n");
        cc.write("  a_.startRecord(*this,tag);\n");
        fIdx = 0;
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            String name = jf.getName();
            if (jf.getType() instanceof JBuffer) {
                cc.write("  { size_t len=0; a_.deserialize("+name+",len,\""+jf.getTag()+"\");}\n");
            } else {
                cc.write("  a_.deserialize("+name+",\""+jf.getTag()+"\");\n");
            }
        }
        cc.write("  a_.endRecord(*this,tag);\n");
        cc.write("  return;\n");
        cc.write("}\n");
        
        
        cc.write("bool "+getCppFQName()+"::operator< (const "+getCppFQName()+"& peer_) const {\n");
        cc.write("  return (1\n");
        for (Iterator i = mFields.iterator(); i.hasNext();) {
            JField jf = (JField) i.next();
            String name = jf.getName();
            cc.write("    && ("+name+" < peer_."+name+")\n");
        }
        cc.write("  );\n");
        cc.write("}\n");
        
        cc.write("bool "+getCppFQName()+"::operator== (const "+getCppFQName()+"& peer_) const {\n");
        cc.write("  return (1\n");
        for (Iterator i = mFields.iterator(); i.hasNext();) {
            JField jf = (JField) i.next();
            String name = jf.getName();
            cc.write("    && ("+name+" == peer_."+name+")\n");
        }
        cc.write("  );\n");
        cc.write("}\n");
        
        cc.write("const ::std::string&"+getCppFQName()+"::type() const {\n");
        cc.write("  static const ::std::string type_(\""+mName+"\");\n");
        cc.write("  return type_;\n");
        cc.write("}\n");
        
        cc.write("const ::std::string&"+getCppFQName()+"::signature() const {\n");
        cc.write("  static const ::std::string sig_(\""+getSignature()+"\");\n");
        cc.write("  return sig_;\n");
        cc.write("}\n");
        
    }
    
    public void genJavaCode(String destDir) throws IOException {
        String pkg = getJavaPackage();
        String pkgpath = pkg.replaceAll("\\.", "/");
        File pkgdir = new File(destDir, pkgpath);
        if (!pkgdir.exists()) {
            // create the pkg directory
            boolean ret = pkgdir.mkdirs();
            if (!ret) {
                throw new IOException("Cannnot create directory: "+pkgpath);
            }
        } else if (!pkgdir.isDirectory()) {
            // not a directory
            throw new IOException(pkgpath+" is not a directory.");
        }
        File jfile = new File(pkgdir, getName()+".java");
        FileWriter jj = new FileWriter(jfile);
        jj.write("// File generated by hadoop record compiler. Do not edit.\n");
        jj.write("package "+getJavaPackage()+";\n\n");
        jj.write("import java.io.IOException;\n");
        jj.write("import org.apache.commons.logging.Log;\n");
        jj.write("import org.apache.commons.logging.LogFactory;\n");
        jj.write("import org.apache.hadoop.io.WritableComparator;\n");
        jj.write("import org.apache.hadoop.io.WritableComparable;\n");
        jj.write("import org.apache.hadoop.io.WritableUtils;\n");
        jj.write("import org.apache.hadoop.io.BytesWritable;\n");
        jj.write("import org.apache.hadoop.io.Text;\n\n");
        jj.write("public class "+getName()+" implements org.apache.hadoop.record.Record, WritableComparable {\n");
        jj.write("  private static final Log LOG= LogFactory.getLog(\""+
            this.getJavaFQName()+"\");\n");
        for (Iterator i = mFields.iterator(); i.hasNext();) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaDecl());
        }
        jj.write("  public "+getName()+"() { }\n");
        
        
        jj.write("  public "+getName()+"(\n");
        int fIdx = 0;
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaConstructorParam(fIdx));
            jj.write((!i.hasNext())?"":",\n");
        }
        jj.write(") {\n");
        fIdx = 0;
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaConstructorSet(fIdx));
        }
        jj.write("  }\n");
        fIdx = 0;
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaGetSet(fIdx));
        }
        jj.write("  public void serialize(org.apache.hadoop.record.OutputArchive a_, String tag) throws java.io.IOException {\n");
        jj.write("    a_.startRecord(this,tag);\n");
        for (Iterator i = mFields.iterator(); i.hasNext();) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaWriteMethodName());
        }
        jj.write("    a_.endRecord(this,tag);\n");
        jj.write("  }\n");
        
        jj.write("  public void deserialize(org.apache.hadoop.record.InputArchive a_, String tag) throws java.io.IOException {\n");
        jj.write("    a_.startRecord(tag);\n");
        for (Iterator i = mFields.iterator(); i.hasNext();) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaReadMethodName());
        }
        jj.write("    a_.endRecord(tag);\n");
        jj.write("}\n");
        
        jj.write("  public String toString() {\n");
        jj.write("    try {\n");
        jj.write("      java.io.ByteArrayOutputStream s =\n");
        jj.write("        new java.io.ByteArrayOutputStream();\n");
        jj.write("      org.apache.hadoop.record.CsvOutputArchive a_ = \n");
        jj.write("        new org.apache.hadoop.record.CsvOutputArchive(s);\n");
        jj.write("      a_.startRecord(this,\"\");\n");
        fIdx = 0;
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaWriteMethodName());
        }
        jj.write("      a_.endRecord(this,\"\");\n");
        jj.write("      return new String(s.toByteArray(), \"UTF-8\");\n");
        jj.write("    } catch (Throwable ex) {\n");
        jj.write("      throw new RuntimeException(ex);\n");
        jj.write("    }\n");
        jj.write("  }\n");
        
        jj.write("  public void write(java.io.DataOutput out) throws java.io.IOException {\n");
        jj.write("    org.apache.hadoop.record.BinaryOutputArchive archive = new org.apache.hadoop.record.BinaryOutputArchive(out);\n");
        jj.write("    serialize(archive, \"\");\n");
        jj.write("  }\n");
        
        jj.write("  public void readFields(java.io.DataInput in) throws java.io.IOException {\n");
        jj.write("    org.apache.hadoop.record.BinaryInputArchive archive = new org.apache.hadoop.record.BinaryInputArchive(in);\n");
        jj.write("    deserialize(archive, \"\");\n");
        jj.write("  }\n");
        
        jj.write("  public int compareTo (Object peer_) throws ClassCastException {\n");
        jj.write("    if (!(peer_ instanceof "+getName()+")) {\n");
        jj.write("      throw new ClassCastException(\"Comparing different types of records.\");\n");
        jj.write("    }\n");
        jj.write("    "+getName()+" peer = ("+getName()+") peer_;\n");
        jj.write("    int ret = 0;\n");
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaCompareTo());
            jj.write("    if (ret != 0) return ret;\n");
        }
        jj.write("     return ret;\n");
        jj.write("  }\n");
        
        jj.write("  public boolean equals(Object peer_) {\n");
        jj.write("    if (!(peer_ instanceof "+getName()+")) {\n");
        jj.write("      return false;\n");
        jj.write("    }\n");
        jj.write("    if (peer_ == this) {\n");
        jj.write("      return true;\n");
        jj.write("    }\n");
        jj.write("    "+getName()+" peer = ("+getName()+") peer_;\n");
        jj.write("    boolean ret = false;\n");
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaEquals());
            jj.write("    if (!ret) return ret;\n");
        }
        jj.write("     return ret;\n");
        jj.write("  }\n");
        
        jj.write("  public int hashCode() {\n");
        jj.write("    int result = 17;\n");
        jj.write("    int ret;\n");
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaHashCode());
            jj.write("    result = 37*result + ret;\n");
        }
        jj.write("    return result;\n");
        jj.write("  }\n");
        jj.write("  public static String signature() {\n");
        jj.write("    return \""+getSignature()+"\";\n");
        jj.write("  }\n");
        
        jj.write("  public static class Comparator extends WritableComparator {\n");
        jj.write("    public Comparator() {\n");
        jj.write("      super("+getName()+".class);\n");
        jj.write("    }\n");

        jj.write("    static public int slurpRaw(byte[] b, int s, int l) {\n");
        jj.write("      try {\n");
        jj.write("        int os = s;\n");
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaSlurpBytes("b","s","l"));
        }
        jj.write("        return (os - s);\n");
        jj.write("      } catch(IOException e) {\n");
        jj.write("        LOG.warn(e);\n");
        jj.write("        throw new RuntimeException(e);\n");
        jj.write("      }\n");
        jj.write("    }\n");
        
        jj.write("    static public int compareRaw(byte[] b1, int s1, int l1,\n");
        jj.write("                       byte[] b2, int s2, int l2) {\n");
        jj.write("      try {\n");
        jj.write("        int os1 = s1;\n");
        for (Iterator i = mFields.iterator(); i.hasNext(); fIdx++) {
            JField jf = (JField) i.next();
            jj.write(jf.genJavaCompareBytes());
        }
        jj.write("        return (os1 - s1);\n");
        jj.write("      } catch(IOException e) {\n");
        jj.write("        LOG.warn(e);\n");
        jj.write("        throw new RuntimeException(e);\n");
        jj.write("      }\n");
        jj.write("    }\n");
        jj.write("    public int compare(byte[] b1, int s1, int l1,\n");
        jj.write("                       byte[] b2, int s2, int l2) {\n");
        jj.write("      int ret = compareRaw(b1,s1,l1,b2,s2,l2);\n");
        jj.write("      return (ret == -1)? -1 : ((ret==0)? 1 : 0);");
        jj.write("    }\n");
        jj.write("  }\n\n");
        jj.write("  static {\n");
        jj.write("    WritableComparator.define("+getName()+".class, new Comparator());\n");
        jj.write("  }\n");

        
        jj.write("}\n");
        
        jj.close();
    }
}
